# -*- coding: utf-8 -*-

# Copyright 2008-2011 Richard Dymond (rjdymond@gmail.com)
#
# This file is part of SkoolKit.
#
# SkoolKit is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# SkoolKit is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# SkoolKit. If not, see <http://www.gnu.org/licenses/>.

import sys
import re
import posixpath
import os.path
from os.path import isfile, isdir, basename
import cgi
try:
    import gd
except ImportError:
    gd = None
try:
    from PIL import Image
except:
    Image = None
from . import version
from skoollib import Udg, MAIN_CODE_ID
from skoolparser import TableParser, RefParser, parse_int, CASE_LOWER

NULL_IMG_LIB = 'null'

# Ref file section prefixes
GRAPHIC_GLITCH = 'GraphicGlitch'
GLOSSARY = 'Glossary'
BUG = 'Bug'
FACT = 'Fact'
POKE = 'Poke'

# Page IDs (as used in the [Paths], [Titles] and [Links] sections)
P_GAME_INDEX = 'GameIndex'
P_MEMORY_MAP = 'MemoryMap'
P_ROUTINES_MAP = 'RoutinesMap'
P_DATA_MAP = 'DataMap'
P_MESSAGES_MAP = 'MessagesMap'
P_GRAPHICS = 'Graphics'
P_GRAPHIC_GLITCHES = 'GraphicGlitches'
P_GSB = 'GameStatusBuffer'
P_MAIN_TIMETABLE = 'MainTimetable'
P_COMMAND_LISTS = 'CommandLists'
P_GLOSSARY = 'Glossary'
P_FACTS = 'Facts'
P_BUGS = 'Bugs'
P_POKES = 'Pokes'

class SkoolWriter:
    def __init__(self, skool=None, file_info=None, case=None, img_lib='gd'):
        self.skool = skool
        self.file_info = file_info
        self.case = case
        self.paths = {}
        self.titles = {}
        self.links = {}
        self.palette = {}
        if skool:
            self.table_parser = TableParser()
            self._create_macros()
            self.paths.update(self._get_paths())
            self.paths.update(self.skool.paths)
            self.titles.update(self._get_titles())
            self.titles.update(self.skool.titles)
            self.links.update(self._get_links())
            self.links.update(self.skool.links)
            self.palette.update(self._get_palette())
            self.palette.update(self.skool.colours)
            if img_lib == NULL_IMG_LIB:
                self.write_image = self._null_write_image
            elif img_lib == 'gd':
                self.write_image = self._gd_write_image
            elif img_lib == 'PIL':
                self.write_image = self._pil_write_image
                self._pil_create_colours()

    def _get_paths(self):
        buffers_path = 'buffers'
        graphics_path = 'graphics'
        images_path = 'images'
        lessons_path = 'lessons'
        maps_path = 'maps'
        reference_path = 'reference'
        astiles_path = join(graphics_path, 'astiles')
        patiles_path = join(graphics_path, 'patiles')
        paths = {
            'FontImagePath': join(images_path, 'font'),
            'JavaScriptPath': '.',
            'LessonsPath': lessons_path,
            'ScreenshotImagePath': join(images_path, 'scr'),
            'StyleSheetPath': '.',
            'UDGImagePath': join(images_path, 'udgs'),
            'Logo': join(images_path, 'logo.png'),
            'StyleSheet': 'style.css',
            P_GAME_INDEX: 'index.html',
            P_MEMORY_MAP: join(maps_path, 'all.html'),
            P_ROUTINES_MAP: join(maps_path, 'routines.html'),
            P_DATA_MAP: join(maps_path, 'data.html'),
            P_MESSAGES_MAP: join(maps_path, 'messages.html'),
            P_BUGS: join(reference_path, 'bugs.html'),
            P_FACTS: join(reference_path, 'facts.html'),
            P_GLOSSARY: join(reference_path, 'glossary.html'),
            P_POKES: join(reference_path, 'pokes.html'),
            P_GSB: join(buffers_path, 'gbuffer.html'),
            P_GRAPHICS: join(graphics_path, 'graphics.html'),
            P_GRAPHIC_GLITCHES: join(graphics_path, 'glitches.html'),
        }
        return paths

    def _get_titles(self):
        titles = {
            P_BUGS: 'Bugs',
            P_DATA_MAP: 'Data',
            P_FACTS: 'Trivia',
            P_GAME_INDEX: 'Index',
            P_GSB: 'Game status buffer',
            P_GLOSSARY: 'Glossary',
            P_GRAPHIC_GLITCHES: 'Graphic glitches',
            P_GRAPHICS: 'Graphics',
            P_MEMORY_MAP: 'Memory map',
            P_MESSAGES_MAP: 'Messages',
            P_POKES: 'Pokes',
            P_ROUTINES_MAP: 'Routines'
        }
        return titles

    def _get_links(self):
        links = {}
        for page_id, text in self._get_titles().items():
            links[page_id] = (text, '')
        links.update({
            P_COMMAND_LISTS: ('Command lists', ''),
            P_MAIN_TIMETABLE: ('Main timetable', ''),
            P_MEMORY_MAP: ('Everything', '')
        })
        return links

    def _get_palette(self):
        palette = {
            'BLACK': (0, 0, 0),
            'BLUE': (0, 0, 197),
            'RED': (197, 0, 0),
            'MAGENTA': (197, 0, 197),
            'GREEN': (0, 198, 0),
            'CYAN': (0, 198, 197),
            'YELLOW': (197, 198, 0),
            'WHITE': (205, 198, 205),
            'BRIGHT_BLACK': (0, 0, 0),
            'BRIGHT_BLUE': (0, 0, 255),
            'BRIGHT_RED': (255, 0, 0),
            'BRIGHT_MAGENTA': (255, 0, 255),
            'BRIGHT_GREEN': (0, 255, 0),
            'BRIGHT_CYAN': (0, 255, 255),
            'BRIGHT_YELLOW': (255, 255, 0),
            'BRIGHT_WHITE': (255, 255, 255),
            'TRANSPARENT': (0, 254, 0)
        }
        return palette

    def has_lessons(self):
        return False

    def has_gbuffer(self):
        return self.skool.has_gbuffer()

    def has_graphics(self):
        return 'Graphics' in self.skool.ref_parser.sections

    def has_graphic_glitches(self):
        return self.skool.has_ref_items(GRAPHIC_GLITCH)

    def has_glossary(self):
        return self.skool.has_ref_items(GLOSSARY)

    def has_bugs(self):
        return self.skool.has_ref_items(BUG)

    def has_facts(self):
        return self.skool.has_ref_items(FACT)

    def has_pokes(self):
        return self.skool.has_ref_items(POKE)

    def get_page_ids(self):
        return self.skool.page_ids

    def need_image(self, fname):
        return self.file_info.need_image(fname)

    def file_exists(self, fname):
        return self.file_info.file_exists(fname)

    def write_asm_entries(self):
        self.write_entries(self.skool.code_path, self.paths[P_MEMORY_MAP])

    def as_img(self, cwd, num, scale=2, mask=1, attr=120, udg_page=None):
        mask_infix = 'm' if mask else 'u'
        udg_page_infix = udg_page if udg_page is not None else ''
        snapshot_name = self.skool.get_snapshot_name()
        snapshot_infix = '_%s' % snapshot_name if snapshot_name else ''
        fname = 'as%03i_%ix%i%s%s%s' % (num, attr, scale, mask_infix, udg_page_infix, snapshot_infix)
        asimg_fname = join('images', 'animatory_states', '%s.png' % fname)
        if self.need_image(asimg_fname):
            self.write_image(self.skool.build_sprite(num, attr, udg_page), asimg_fname, scale=scale, mask=mask)
        img_path = FileInfo.relpath(cwd, asimg_fname)
        return '<img alt="Animatory state %i" src="%s" />' % (num, img_path)

    def play_area(self, cwd, fname, x, y, w=1, h=1, scale=2, show_chars=0, show_x=0):
        img_fname = self.get_img_fname(fname, 'images/play_area')
        if self.need_image(img_fname):
            self.write_image(self.skool.get_skool_udgs(x, y, w, h, show_chars, show_x), img_fname, scale=scale)
        img_path = FileInfo.relpath(cwd, img_fname)
        return '<img alt="%s" src="%s" />' % (fname, img_path)

    def write_playarea_tile_table(self, cwd, tiles, id_prefix='s'):
        lines = []
        lines.append('<table class="tiles">\n')
        for row in tiles:
            lines.append('<tr>\n')
            for tile in row:
                lines.append('<td>')
                if tile:
                    bubble_id = '%s%ix%i' % (id_prefix, tile.x, tile.y)
                    title = '(x,y)=(%i,%i)' % (tile.x, tile.y)
                    img_fname = '%x%x.png' % (tile.ref + 256 * tile.udg_page, tile.attr)
                    img_path = join(cwd, img_fname)
                    if self.need_image(img_path):
                        self.write_image([[tile]], img_path, scale=4)
                    img = '<img alt="%i,%i" src="%s"/>' % (tile.x, tile.y, img_fname)
                    lines.append('<div class="T" onclick="s(\'%s\')">%s</div>' % (bubble_id, img))
                    lines += self._get_bubble(bubble_id, img, title, '<br/>'.join(self.get_playarea_tile_text(tile)))
                lines.append('</td>\n')
            lines.append('</tr>\n')
        lines.append('</table>\n')
        return lines

    def write_mutable_tile_tables(self, cwd):
        return []

    def play_area_tiles(self, cwd, w, h):
        lines = []
        lines += self.write_playarea_tile_table(cwd, self.skool.get_skool_udgs(0, 0, w, h))
        lines += self.write_mutable_tile_tables(cwd)
        return ''.join(lines)

    def _animatory_state_row(self, cwd, rownum):
        lines = []
        for state in (rownum, rownum + 128):
            lines.append('<td>')
            lines.append('<a name="%i" />%s' % (state, self.as_img(cwd, state)))
            lines.append('</td>')
        lines.append('<td>%i, %i: %s</td>' % (rownum, rownum + 128, self.skool.as_descs.get(rownum, '-')))
        return lines

    def animatory_states(self, cwd):
        lines = ['<table>']
        for n in range(128):
            lines.append('<tr>')
            lines += self._animatory_state_row(cwd, n)
            lines.append('</tr>')
        lines.append('</table>')
        return '\n'.join(lines)

    def _get_animatory_state_tiles_row(self, n):
        states = ((n, None),)
        states_desc = '%i: %s' % (n, self.skool.as_descs[n])
        return states, states_desc

    def _get_bubble(self, bubble_id, img, title, text):
        return [
            '<div id="%s" class="B P" onclick="h(this)">' % bubble_id,
            '<div class="H">%s</div>' % title,
            '<table><tr><td>%s</td><td>%s</td></tr></table>' % (img, text),
            '</div>'
        ]

    def astiles(self, cwd):
        lines = []
        attr = 120
        lines.append('<table>\n')
        for n in range(128):
            row = []
            lines.append('<tr>\n')
            states, states_desc = self._get_animatory_state_tiles_row(n)
            for state, udg_page in states:
                lines.append('<td>\n')
                lines.append('<table class="A">\n')
                sprite = self.skool.build_sprite(state, attr, udg_page)
                num_rows = len(sprite)
                for row_num in range(num_rows):
                    lines.append('<tr>\n')
                    row = sprite[row_num]
                    for col_num in range(len(row)):
                        tile = row[col_num]
                        bubble_id_suffix = '%02x' % udg_page if udg_page is not None else ''
                        bubble_id = 'B%02x%x%s' % (state, row_num + num_rows * col_num, bubble_id_suffix)
                        title = '%i: row %i, column %i' % (state, row_num, col_num)
                        suffix = ',%i' % udg_page if udg_page is not None else ''
                        img_fname = '%x.png' % (tile.ref + 256 * tile.udg_page)
                        img_path = join(cwd, img_fname)
                        if self.need_image(img_path):
                            self.write_image([[tile]], img_path, scale=4, mask=1)
                        img = '<img alt="%i:%i,%i" src="%s"/>' % (state, row_num, col_num, img_fname)
                        text = ['<i>lsb</i>=%i, <i>ref_page</i>=%i' % (tile.ref_addr % 256, tile.ref_addr // 256)]
                        text.append('<i>ref_addr</i>=(<i>lsb</i>,<i>ref_page</i>)=%i' % tile.ref_addr)
                        if tile.ref == 0:
                            text.append('<i>ref</i>=%i' % tile.ref)
                        else:
                            text.append('<i>ref</i>=%i, <i>udg_page</i>=%i' % (tile.ref, tile.udg_page))
                            text.append('<i>udg_addr</i>=(<i>ref</i>,<i>udg_page</i>)=%i' % (tile.ref + 256 * tile.udg_page))
                        lines.append('<td>')
                        lines.append('<div class="T" onclick="s(\'%s\')">%s</div>' % (bubble_id, img))
                        lines += self._get_bubble(bubble_id, img, title, '<br/>'.join(text))
                        lines.append('</td>\n')
                    lines.append('</tr>\n')
                lines.append('</table>\n')
                lines.append('</td>\n')
            lines.append('<td>%s</td>\n' % states_desc)
            lines.append('</tr>\n')
        lines.append('</table>\n')
        return ''.join(lines)

    def write_graphics(self):
        ofile, cwd = self.open_file(self.paths[P_GRAPHICS])
        self.write_header(ofile, self.titles[P_GRAPHICS], cwd, 'graphics')
        ofile.write(self.expand(self.skool.get_section('Graphics'), cwd))
        ofile.write('\n')
        self.write_footer(ofile)

    def write_graphic_glitches(self):
        self.write_box_page(self.paths[P_GRAPHIC_GLITCHES], self.titles[P_GRAPHIC_GLITCHES], 'graphics', self.skool.get_ref_items(GRAPHIC_GLITCH))

    def write_index(self):
        ofile, cwd = self.open_file(self.paths[P_GAME_INDEX])
        self.write_head(ofile, 'Index', cwd)
        ofile.write('<body class="main">\n')
        table = Table("header")
        row = []
        prefix = self.skool.game_vars.get('TitlePrefix', 'The complete')
        suffix = self.skool.game_vars.get('TitleSuffix', 'RAM disassembly')
        row.append(Cell(prefix, "headerText"))
        row.append(Cell(self._get_logo(cwd)))
        row.append(Cell(suffix, "headerText"))
        table.add_row(row)
        ofile.write(table.html())

        link_groups = {
            'MemoryMaps': ('Memory maps', (
                P_MEMORY_MAP,
                P_ROUTINES_MAP,
                P_DATA_MAP,
                P_MESSAGES_MAP
            )),
            'Graphics': ('Graphics', (
                P_GRAPHICS,
                P_GRAPHIC_GLITCHES
            )),
            'DataTables': ('Data tables and buffers', (
                P_GSB
            )),
            'Reference': ('Reference', (
                P_GLOSSARY,
                P_FACTS,
                P_BUGS,
                P_POKES
            ))
        }
        for section_id, header_text, page_list in self.skool.get_ref_items('Index', as_list=True):
            link_groups[section_id] = (header_text, page_list)
        sections = {}
        for section_id, (header_text, page_list) in link_groups.items():
            links = []
            for page_id in page_list:
                fname = self.paths.get(page_id)
                if fname and self.file_exists(fname):
                    link_file = FileInfo.relpath(cwd, fname)
                    link_text = self.links[page_id]
                    links.append((link_file, link_text[0], link_text[1]))
            sections[section_id] = ('<div class="headerText">%s</div>\n' % header_text, links)
        other_code_links = []
        for code_id, code in self.skool.other_code:
            fname = code['Index']
            if self.file_exists(fname):
                link_file = FileInfo.relpath(cwd, fname)
                other_code_links.append((link_file, code['Title'], ''))
        sections['OtherCode'] = ('<div class="headerText">Other code</div>\n', other_code_links)

        default_index = ('MemoryMaps', 'Graphics', 'DataTables', 'OtherCode', 'Reference')
        for section_id in self.skool.get_list('Index', default_index):
            header, links = sections.get(section_id, ('', ()))
            if links:
                ofile.write(header)
                ofile.write('<ul class="indexList">\n')
                for href, link_text, other_text in links:
                    ofile.write('  <li><a class="link" href="%s">%s</a>%s</li>\n' % (href, link_text, other_text))
                ofile.write('</ul>\n')

        self.write_footer(ofile)

    def _get_pt_link(self, cwd, char_num):
        pt_address = self.skool.min_lesson + 256 * char_num
        pt_file = FileInfo.asm_relpath(cwd, pt_address, self.skool.get_code_path(MAIN_CODE_ID))
        return '<a class="link" href="%s">%i</a>' % (pt_file, char_num)

    def lesson_index(self, cwd):
        li_table = Table('lesson', 'ID', 'Lesson')
        for i in range(self.skool.min_lesson, self.skool.max_lesson + 1):
            row = []
            lesson_file = FileInfo.relpath(cwd, self.paths['Lesson%i' % i])
            row.append(Cell('<a class="link" href="%s">%i</a>' % (lesson_file, i), "lessonType"))
            row.append(Cell(self.skool.lesson_descs[i], "lesson"))
            li_table.add_row(row)
        return li_table.html()

    def _lesson_prev_next(self, prev, next, up):
        prev_text = next_text = up_text = ''
        if prev:
            prev_text = 'Prev: %s' % prev
        if next:
            next_text = 'Next: %s' % next
        if up:
            up_text = 'Up: %s' % up
        lines = ['<table class="prevNext">', '<tr>']
        for td_class, text in (('prev', prev_text), ('up', up_text), ('next', next_text)):
            if text:
                lines.append('<td class="%s">%s</td>' % (td_class, text))
            else:
                lines.append('<td class="%s"/>' % td_class)
        lines.append('</tr>')
        lines.append('</table>')
        return lines

    def lesson(self, cwd, lesson):
        lines = []
        index_path = FileInfo.relpath(cwd, self.paths['LessonIndex'])

        prev = None if lesson == self.skool.min_lesson else self.expand('#LINK:Lesson%i(%i)' % (lesson - 1, lesson - 1), cwd)
        up = '<a class="link" href="%s">Index</a>' % index_path
        next = None if lesson == self.skool.max_lesson else self.expand('#LINK:Lesson%i(%i)' % (lesson + 1, lesson + 1), cwd)
        prev_up_next = self._lesson_prev_next(prev, next, up)

        lines += prev_up_next
        lines.append('<div class="description">Lesson %i: %s</div>' % (lesson, self.skool.lesson_descs[lesson]))
        lines.append('<table class="lesson">')
        lines.append('<tr>')
        lines.append('<th>Character</th>')
        lines.append('<th>Address</th>')
        lines.append('<th>Command list</th>')
        lines.append('</tr>')
        for c in range(self.skool.min_character, self.skool.max_character + 1):
            lesson_address = lesson + 256 * c
            tap = self.skool.snapshot[lesson_address]
            pt_link = self._get_pt_link(cwd, c)
            char_desc = self.skool.characters[c]
            tap_address = self.skool.tap_addresses[tap]
            tap_link = self.expand('#R%i(%i)' % (tap_address, tap), cwd)
            tap_desc = self.skool.tap_descs[tap]
            lines.append('<tr>')
            lines.append('<td class="lessonCharacter">%s - %s</td>' % (pt_link, char_desc))
            lines.append('<td class="lessonAddress">%i</td>' % lesson_address)
            lines.append('<td class="lesson">%s - %s</td>' % (tap_link, tap_desc))
            lines.append('</tr>')
        lines.append('</table>')
        lines += prev_up_next

        return '\n'.join(lines)

    def personal_timetables(self, cwd):
        pt_table = Table('timetable', 'No.', 'Character')
        for c in range(self.skool.min_character, self.skool.max_character + 1):
            row = []
            row.append(Cell(self._get_pt_link(cwd, c)))
            row.append(Cell(self.skool.characters[c]))
            pt_table.add_row(row)
        return pt_table.html()

    def write_memory_map(self):
        self.write_map(self.paths[P_MEMORY_MAP], self.titles[P_MEMORY_MAP], '', self.skool.code_path, True)

    def write_routines_map(self):
        self.write_map(self.paths[P_ROUTINES_MAP], self.titles[P_ROUTINES_MAP], 'r', self.skool.code_path)

    def write_data_map(self):
        self.write_map(self.paths[P_DATA_MAP], self.titles[P_DATA_MAP], 'd', self.skool.code_path, True)

    def write_messages_map(self):
        self.write_map(self.paths[P_MESSAGES_MAP], self.titles[P_MESSAGES_MAP], 't', self.skool.code_path)

    def _characters_table(self, cwd, char_buffer_base_byte):
        characters = self.skool.characters
        if not characters:
            return ''
        first_char_num = min(characters.keys())

        cols = 3

        headers = []
        for c in range(cols):
            headers.append(Header('Page'))
            headers.append(Header('Address'))
            headers.append(Header('Character'))
            if c < cols - 1:
                headers.append(Cell('', "transparent"))
        char_table = Table('cbuffer')
        char_table.add_row(headers)

        rows = (cols - 1  + len(characters)) // cols
        for r in range(rows):
            row = []
            for c in range(cols):
                char_index = r + c * rows
                if char_index < len(characters):
                    char_num = first_char_num + char_index
                    row.append(Cell(char_num, "centre"))
                    buf_addr = char_buffer_base_byte + char_num * 256
                    entry_link = self.expand('#R%i' % buf_addr, cwd)
                    row.append(Cell(entry_link, "centre"))
                    row.append(Cell(characters[char_num]))
                    if c < cols - 1:
                        row.append(Cell('', "transparent"))
            char_table.add_row(row)

        return char_table.html()

    def character_buffers(self, cwd, char_buffer_base_byte):
        lines = [self._characters_table(cwd, char_buffer_base_byte)]
        lines.append('<p>The bytes in each character buffer are used as follows:</p>\n')
        cb_table = Table('cbuffer', 'Byte(s)', 'Contents')
        for bytes, entries in self.skool.char_buf_descs:
            rows = len(entries)
            anchors = ''.join(['<a name="%s" />' % name for name in bytes.split(',')])
            row = []
            row.append(Cell(bytes, "centre", rowspan=rows))
            row.append(Cell("%s%s" % (anchors, self.expand(entries[0], cwd)), "byteDesc"))
            cb_table.add_row(row)
            for i in range(1, rows):
                cb_table.add_row((Cell(self.expand(entries[i], cwd), "byteDesc"),))
        lines.append(cb_table.html())
        return ''.join(lines)

    def write_gbuffer(self):
        ofile, cwd = self.open_file(self.paths[P_GSB])
        self.write_header(ofile, self.titles[P_GSB], cwd, "gbuffer")
        gb_table = Table("gbuffer", 'Address', 'Length', 'Purpose')
        gsb_includes = self.skool.game_vars.get('GameStatusBufferIncludes', [])
        if gsb_includes:
            gsb_includes = [parse_int(a) for a in gsb_includes.split(',')]
        for entry in self.skool.memory_map:
            if not (entry.ctl == 'g' or entry.address in gsb_includes):
                continue
            address = entry.address
            desc = '    <div class="gbufDesc">%s</div>' % entry.description
            if entry.details:
                desc += '\n    <div class="gbufDetails">%s</div>' % entry.details
            row = []
            asm_file = FileInfo.asm_relpath(cwd, address, self.skool.code_path)
            entry_link = '<a name="%i" class="link" href="%s">%s</a>' % (address, asm_file, entry.addr_str)
            row.append(Cell(entry_link, "bufAddress"))
            row.append(Cell(entry.size, "bufLength"))
            row.append(Cell("\n%s\n" % self.expand(desc, cwd), "bufDesc"))
            gb_table.add_row(row)
        ofile.write(gb_table.html())
        self.write_footer(ofile)

    def write_glossary(self):
        ofile, cwd = self.open_file(self.paths[P_GLOSSARY])
        self.write_header(ofile, self.titles[P_GLOSSARY], cwd, 'glossary')
        table = Table('glossary', 'Term', 'Description')
        for term, paragraphs in self.skool.get_ref_items(GLOSSARY):
            row = []
            row.append(Cell(term, "glossaryTerm"))
            row.append(Cell(self.expand(paragraphs[0], cwd), "glossaryDefn"))
            table.add_row(row)
        ofile.write(table.html())
        self.write_footer(ofile)

    def write_link_list(self, ofile, link_list):
        ofile.write('<ul class="linkList">\n')
        for anchor, title in link_list:
            ofile.write('<li><a class="link" href="#%s">%s</a></li>\n' % (anchor, title))
        ofile.write('</ul>\n')

    def write_boxes(self, ofile, boxes, cwd):
        box_num = 1
        for anchor, title, paragraphs in boxes:
            ofile.write('<div><a name="%s"></a></div>\n' % anchor)
            ofile.write('<div class="box box%i">\n' % box_num)
            ofile.write('<div class="boxTitle">%s</div>\n' % title)
            for paragraph in paragraphs:
                ofile.write('<div class="paragraph">%s</div>\n' % self.expand(paragraph, cwd))
            ofile.write('</div>\n')
            box_num = 3 - box_num

    def write_box_page(self, fname, title, body_class, boxes):
        ofile, cwd = self.open_file(fname)
        self.write_header(ofile, title, cwd, body_class)
        self.write_link_list(ofile, [(anchor, title) for anchor, title, p in boxes])
        self.write_boxes(ofile, boxes, cwd)
        self.write_footer(ofile)

    def write_pokes(self):
        self.write_box_page(self.paths[P_POKES], self.titles[P_POKES], 'pokes', self.skool.get_ref_items(POKE))

    def write_bugs(self):
        self.write_box_page(self.paths[P_BUGS], self.titles[P_BUGS], 'bugs', self.skool.get_ref_items(BUG))

    def write_facts(self):
        self.write_box_page(self.paths[P_FACTS], self.titles[P_FACTS], 'facts', self.skool.get_ref_items(FACT))

    def write_prev_next(self, ofile, prev, next, up):
        if prev or next or up:
            prev_text = next_text = up_text = ''
            if prev:
                prev_text = 'Prev: %s' % prev
            if next:
                next_text = 'Next: %s' % next
            if up:
                up_text = 'Up: %s' % up
            pn_table = Table("prevNext")
            pn_table.add_row((Cell(prev_text, "prev"), Cell(up_text, "up"), Cell(next_text, "next")))
            ofile.write(pn_table.html())

    def write_entry(self, cwd, entry, fname=None, page_header=None, prev=None, next=None, up=None):
        fname = fname or FileInfo.asm_fname(entry.address)
        ofile = self.file_info.open_file(cwd, fname)
        body_class = "disassembly"
        if entry.is_routine():
            title_format = "Routine at %s"
            page_header = page_header or "Routines"
        elif entry.ctl in 'uz':
            title_format = "Unused RAM at %s"
            page_header = page_header or "Unused"
        elif entry.ctl == 'g':
            title_format = "Game status buffer entry at %s"
            page_header = page_header or "Game status buffer"
        else:
            title_format = "Data at %s"
            page_header = page_header or "Data"
        self.write_header(ofile, title_format % entry.addr_str, cwd, body_class, page_header)

        if prev:
            prev_file = FileInfo.asm_fname(prev.address)
            prev = '<a class="link" href="%s">%s</a>' % (prev_file, prev.addr_str)
        if next:
            next_file = FileInfo.asm_fname(next.address)
            next = '<a class="link" href="%s">%s</a>' % (next_file, next.addr_str)
        if up:
            map_path = FileInfo.relpath(cwd, up)
            up = '<a class="link" href="%s#%s">Map</a>' % (map_path, entry.address)
        self.write_prev_next(ofile, prev, next, up)

        desc = self.expand(entry.description, cwd)
        ofile.write('<div class="description">%s: %s</div>\n' % (entry.addr_str, desc))

        table_class = 'disassembly'
        comment_class = 'comment'
        transparent_class = 'transparentComment'
        if entry.ctl not in 'cuz':
            table_class = 'dataDisassembly'
            comment_class = 'dataComment'
            transparent_class = 'transparentDataComment'
        ofile.write('<table class="%s">\n' % table_class)
        details = self.expand(entry.details, cwd).strip()
        if details or entry.registers:
            ofile.write('<tr>\n')
            ofile.write('<td class="routineComment" colspan="3">\n')
            if details:
                ofile.write('<div class="details">\n%s\n</div>\n' % details)
            if entry.registers:
                ofile.write('<table class="registers">\n')
                for reg in entry.registers:
                    ofile.write('<tr>\n')
                    ofile.write('<td class="register">%s</td>\n' % reg.name)
                    ofile.write('<td class="registerContents">%s</td>\n' % self.expand(reg.contents, cwd))
                    ofile.write('</tr>\n')
                ofile.write('</table>\n')
            ofile.write('</td>\n')
            ofile.write('</tr>\n')
        show_comment_col = False
        for instruction in entry.instructions:
            if instruction.comment and instruction.comment.text:
                show_comment_col = True
        for instruction in entry.instructions:
            mid_routine_comments = self.expand(entry.get_mid_routine_comment(instruction.label), cwd)
            address = instruction.address
            anchor = '<a name="%s"></a>' % address
            if mid_routine_comments:
                ofile.write('<tr>\n')
                ofile.write('<td class="routineComment" colspan="3">\n')
                ofile.write('%s\n' % anchor)
                ofile.write('<div class="comments">%s</div>\n' % mid_routine_comments)
                ofile.write('</td>\n')
                ofile.write('</tr>\n')
                anchor = ''
            ofile.write('<tr>\n')
            tdclass = 'label' if instruction.ctl in 'c*!' else 'address'
            ofile.write('<td class="%s">%s%s</td>\n' % (tdclass, anchor, instruction.addr_str))
            operation, reference = instruction.operation, instruction.reference
            if reference:
                entry_address = reference.entry.address
                name = '' if reference.address == entry_address else '#%s' % reference.address
                if reference.entry.asm_id:
                    # This is a reference to an entry in another disassembly
                    entry_file = FileInfo.asm_relpath(cwd, entry_address, self.skool.get_code_path(reference.entry.asm_id))
                else:
                    # This is a reference to an entry in the same disassembly
                    entry_file = FileInfo.asm_fname(entry_address)
                link = '<a class="link" href="%s%s">%s</a>' % (entry_file, name, reference.addr_str)
                operation = operation.replace(reference.addr_str, link)
            ofile.write('<td class="instruction">%s</td>\n' % operation)
            if show_comment_col:
                comment = instruction.comment
                if comment:
                    rowspan = '' if comment.rowspan < 2 else ' rowspan="%i"' % comment.rowspan
                    if comment.text.startswith('[') and comment.text.endswith(']'):
                        comment_html = comment.text
                    else:
                        comment_html = self.expand(comment.text, cwd)
                    ofile.write('<td class="%s"%s>%s</td>\n' % (comment_class, rowspan, comment_html))
            else:
                ofile.write('<td class="%s" />\n' % transparent_class)
            ofile.write('</tr>\n')

        end_routine_comments = self.expand(entry.end_comment, cwd)
        if end_routine_comments:
            ofile.write('<tr>\n')
            ofile.write('<td class="routineComment" colspan="3">\n')
            ofile.write('<div class="comments">%s</div>\n' % end_routine_comments)
            ofile.write('</td>\n')
            ofile.write('</tr>\n')

        ofile.write('</table>\n')

        self.write_prev_next(ofile, prev, next, up)
        self.write_footer(ofile)

    def write_entries(self, cwd, map_file, page_header=None):
        prev = None
        memory_map = self.skool.memory_map
        for i, entry in enumerate(memory_map):
            next = memory_map[i + 1] if i + 1 < len(memory_map) else None
            self.write_entry(cwd, entry, page_header=page_header, prev=prev, next=next, up=map_file)
            prev = entry

    def write_map(self, map_file, title, entry_types, asm_path, show_page_byte=False):
        if entry_types:
            show_routines = 'r' in entry_types
            show_data = 'd' in entry_types
            show_text = 't' in entry_types
            show_gbuffer = 'g' in entry_types
            show_unused = 'u' in entry_types
        else:
            show_routines = show_data = show_text = show_gbuffer = show_unused = True
        cwd = os.path.dirname(map_file)

        ofile = self.file_info.open_file(map_file)
        self.write_header(ofile, title, cwd, "map")
        details = self.skool.map_details.get(entry_types)
        if details:
            ofile.write('<div class="details">%s</div>\n' % self.expand(details, cwd))
        headers = ['Address', 'Description']
        if show_page_byte:
            headers.insert(0, 'Page')
            headers.insert(1, 'Byte')
        table = Table("map", *headers)

        for entry in self.skool.memory_map:
            purpose = entry.description
            if entry.ctl == 'c' and show_routines:
                entry_class = 'routine'
                desc_class = 'purpose'
            elif entry.ctl in 'bw' and show_data:
                entry_class = 'data'
                desc_class = 'dataDesc'
            elif entry.ctl in 'g' and show_gbuffer:
                entry_class = 'gbuffer'
                desc_class = 'gbufferDesc'
            elif entry.ctl in 'uz' and show_unused:
                size = entry.size
                purpose = 'Unused (%s byte%s)' % (size, 's' if size > 1 else '')
                entry_class = 'unused'
                desc_class = 'unusedDesc'
            elif entry.ctl == 't' and show_text:
                entry_class = 'messageAddress'
                desc_class = 'message'
            else:
                continue

            address = entry.address
            row = []
            if show_page_byte:
                row.append(Cell(address // 256, "centre"))
                row.append(Cell(address % 256, "centre"))
            asm_relpath = FileInfo.relpath(cwd, asm_path)
            asm_file = FileInfo.asm_fname(address, asm_relpath)
            content = '<a class="link" name="%s" href="%s">%s</a>' % (address, asm_file, entry.addr_str)
            row.append(Cell(content, entry_class))
            row.append(Cell(self.expand(purpose, cwd), desc_class))
            table.add_row(row)

        ofile.write(table.html())
        self.write_footer(ofile)

    def write_page(self, page_id):
        ofile, cwd = self.open_file(self.paths[page_id])
        page = self.skool.pages[page_id]
        body_class = page.get('BodyClass')
        js = page.get('JavaScript', 0)
        self.write_header(ofile, self.titles[page_id], cwd, body_class, js=js)
        page_content = self.expand(self.skool.page_contents[page_id], cwd).strip()
        ofile.write('%s\n' % page_content)
        self.write_footer(ofile)

    def open_file(self, fname):
        ofile = self.file_info.open_file(fname)
        cwd = os.path.dirname(fname)
        return ofile, cwd

    def write_prologue(self, ofile):
        ofile.write('<?xml version="1.0" encoding="utf-8" ?>\n')
        ofile.write('<!DOCTYPE html\n')
        ofile.write('    PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN"\n')
        ofile.write('    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">\n')
        ofile.write('<html xmlns="http://www.w3.org/1999/xhtml">\n')

    def write_head(self, ofile, title, cwd, js=False):
        self.write_prologue(ofile)
        ofile.write('<head>\n')
        ofile.write('<title>%s: %s</title>\n' % (self.skool.game, title))
        ofile.write('<link rel="stylesheet" type="text/css" href="%s" />\n' % FileInfo.relpath(cwd, join(self.paths['StyleSheetPath'], self.paths['StyleSheet'])))
        if js:
            ofile.write('<script type="text/javascript" src="%s"></script>\n' % FileInfo.relpath(cwd, join(self.paths['JavaScriptPath'], self.paths['JavaScript'])))
        ofile.write('</head>\n')

    def _get_logo(self, cwd):
        if self.file_exists(self.paths['Logo']):
            return '<img src="%s" alt="%s" />' % (FileInfo.relpath(cwd, self.paths['Logo']), self.skool.game)
        return self.skool.game

    def write_header(self, ofile, title, cwd, body_class, body_title=None, js=False):
        self.write_head(ofile, title, cwd, js)
        index = FileInfo.relpath(cwd, self.paths[P_GAME_INDEX])
        if body_class:
            ofile.write('<body class="%s">\n' % body_class)
        else:
            ofile.write('<body>\n')
        table = Table("header")
        row = []
        row.append(Cell('<a class="link" href="%s">%s</a>' % (index, self._get_logo(cwd))))
        row.append(Cell(body_title or title, "headerText"))
        table.add_row(row)
        ofile.write(table.html())

    def write_footer(self, ofile):
        ofile.write('<div class="footer">\n')
        release_name = self.skool.info.get('Release', '')
        copyright_message = self.skool.info.get('Copyright', '')
        created_message = self.skool.info.get('Created', '')
        if release_name:
            ofile.write('<div>%s</div>\n' % release_name)
        if copyright_message:
            ofile.write('<div>%s</div>\n' % copyright_message)
        if created_message:
            ofile.write('<div>%s</div>\n' % created_message.replace('$VERSION', version))
        ofile.write('</div>\n')
        ofile.write('</body>\n')
        ofile.write('</html>\n')
        ofile.close()

    def write_logo_image(self, cwd):
        logo_macro = self.skool.game_vars.get('Logo')
        if logo_macro:
            self.expand(logo_macro, cwd)
        return logo_macro

    def get_pixel_array(self, udg_array, mask):
        # Convert UDG array into byte array
        byte_array = []
        for row in udg_array:
            for line in range(8):
                byte_array.append([(udg.ink, udg.paper, udg.attr, udg.data[line], udg.mask[line] if mask else 0) for udg in row])

        # Convert byte array into pixel array
        pixel_array = []
        for row in byte_array:
            pixel_array.append([])
            for ink, paper, attr, byte, mask_byte in row:
                bright = (attr & 64) // 8
                ink = ink or (bright + (attr & 7))
                paper = paper or (bright + (attr & 56) // 8)
                for b in range(8):
                    colour = paper
                    pixel = byte & 128
                    if mask:
                        if pixel & mask_byte:
                            colour = ink
                        elif mask_byte & 128:
                            colour = 16
                        mask_byte *= 2
                    elif pixel:
                        colour = ink
                    pixel_array[-1].append(colour)
                    byte *= 2

        return pixel_array

    def _gd_create_colours(self, image):
        colours = []
        colours.append(image.colorAllocate(self.palette['BLACK']))
        colours.append(image.colorAllocate(self.palette['BLUE']))
        colours.append(image.colorAllocate(self.palette['RED']))
        colours.append(image.colorAllocate(self.palette['MAGENTA']))
        colours.append(image.colorAllocate(self.palette['GREEN']))
        colours.append(image.colorAllocate(self.palette['CYAN']))
        colours.append(image.colorAllocate(self.palette['YELLOW']))
        colours.append(image.colorAllocate(self.palette['WHITE']))
        colours.append(colours[0])
        colours.append(image.colorAllocate(self.palette['BRIGHT_BLUE']))
        colours.append(image.colorAllocate(self.palette['BRIGHT_RED']))
        colours.append(image.colorAllocate(self.palette['BRIGHT_MAGENTA']))
        colours.append(image.colorAllocate(self.palette['BRIGHT_GREEN']))
        colours.append(image.colorAllocate(self.palette['BRIGHT_CYAN']))
        colours.append(image.colorAllocate(self.palette['BRIGHT_YELLOW']))
        colours.append(image.colorAllocate(self.palette['BRIGHT_WHITE']))
        colours.append(image.colorAllocate(self.palette['TRANSPARENT']))
        return colours

    def _gd_write_image(self, udgs, img_file, scale=2, mask=False, width=None):
        pixel_array = self.get_pixel_array(udgs, mask)
        width = width or len(pixel_array[0])
        image = gd.image((scale * width, scale * len(pixel_array)))
        colours = self._gd_create_colours(image)
        for y, row in enumerate(pixel_array):
            for x, attr in enumerate(row[:width]):
                x1, y1 = x * scale, y * scale
                x2, y2 = x1 + scale - 1, y1 + scale - 1
                image.filledRectangle((x1, y1), (x2, y2), colours[attr])
        f = self.file_info.open_file(img_file)
        image.writePng(f)
        f.close()

    def _pil_create_colours(self):
        self.colours = []
        self.colours.extend(self.palette['BLACK'])
        self.colours.extend(self.palette['BLUE'])
        self.colours.extend(self.palette['RED'])
        self.colours.extend(self.palette['MAGENTA'])
        self.colours.extend(self.palette['GREEN'])
        self.colours.extend(self.palette['CYAN'])
        self.colours.extend(self.palette['YELLOW'])
        self.colours.extend(self.palette['WHITE'])
        self.colours.extend(self.palette['BRIGHT_BLACK'])
        self.colours.extend(self.palette['BRIGHT_BLUE'])
        self.colours.extend(self.palette['BRIGHT_RED'])
        self.colours.extend(self.palette['BRIGHT_MAGENTA'])
        self.colours.extend(self.palette['BRIGHT_GREEN'])
        self.colours.extend(self.palette['BRIGHT_CYAN'])
        self.colours.extend(self.palette['BRIGHT_YELLOW'])
        self.colours.extend(self.palette['BRIGHT_WHITE'])
        self.colours.extend(self.palette['TRANSPARENT'])

    def _pil_write_image(self, udgs, img_file, scale=2, mask=False, width=None):
        pixel_array = self.get_pixel_array(udgs, mask)
        width = width or len(pixel_array[0])
        image = Image.new('P', (scale * width, scale * len(pixel_array)), None)
        image.putpalette(self.colours)
        pixels = []
        for row in pixel_array:
            pixel_row = []
            for pixel in row[:width]:
                pixel_row.extend([pixel] * scale)
            for i in range(scale):
                pixels.extend(pixel_row)
        image.putdata(pixels)
        f = self.file_info.open_file(img_file)
        image.save(f, 'PNG', optimize=True)
        f.close()

    def _null_write_image(self, udgs, img_file, scale=2, mask=False, width=None):
        f = self.file_info.open_file(img_file)
        f.close()

    def write_image(self, udgs, img_file, scale=2, mask=False, width=None):
        return

    def _create_macros(self):
        self.macros = {}
        self.macros['#BUG'] = self.expand_bug
        self.macros['#C'] = self.expand_c
        self.macros['#CALL'] = self.expand_call
        self.macros['#D'] = self.expand_desc
        self.macros['#EREFS'] = self.expand_erefs
        self.macros['#FACT'] = self.expand_fact
        self.macros['#FONT'] = self.expand_font
        self.macros['#LINK'] = self.expand_link
        self.macros['#NAME'] = self.expand_name
        self.macros['#P'] = self.expand_p
        self.macros['#POKE'] = self.expand_poke
        self.macros['#POKES'] = self.expand_pokes
        self.macros['#POPS'] = self.expand_pops
        self.macros['#PUSHS'] = self.expand_pushs
        self.macros['#R'] = self.expand_reference
        self.macros['#REFS'] = self.expand_refs
        self.macros['#REG'] = self.expand_reg
        self.macros['#SCR'] = self.expand_scr
        self.macros['#SPACE'] = self.expand_space
        self.macros['#TABLE'] = self.expand_table
        self.macros['#UDG'] = self.expand_udg
        self.macros['#UDGARRAY'] = self.expand_udgarray
        self.macros['#UDGTABLE'] = self.expand_table

    def error(self, msg):
        sys.stderr.write('\n*** Error: %s\n' % msg)
        sys.exit(1)

    def build_table(self, table):
        table_class = table.table_class
        html = '\n<table%s>\n' % (' class="%s"' % table_class if table_class else '')
        for row in table.rows:
            html += '<tr>\n'
            for cell in row:
                colspan = ' colspan="%i"' % cell.colspan if cell.colspan > 1 else ''
                rowspan = ' rowspan="%i"' % cell.rowspan if cell.rowspan > 1 else ''
                if cell.header:
                    tag = 'th'
                    td_class = ''
                else:
                    tag = 'td'
                    cell_class = cell.cell_class
                    if cell.transparent:
                        cell_class += " transparent"
                    cell_class = cell_class.lstrip()
                    td_class = ' class="%s"' % cell_class if cell_class else ''
                html += '<%s%s%s%s>%s</%s>\n' % (tag, td_class, colspan, rowspan, cell.contents, tag)
            html += '</tr>\n'
        return html + '</table>\n'

    def _get_params(self, param_string, num=0, req=0, defaults=(), ints_only=False):
        params = []
        if param_string:
            for p in param_string.split(','):
                if p:
                    param = parse_int(p)
                    if param is None:
                        if ints_only:
                            self.error("Cannot parse integer '%s' in macro parameter list: '%s'" % (p, param_string))
                        else:
                            params.append(p)
                    else:
                        params.append(param)
                else:
                    params.append(None)
        params += [None] * (num - len(params))
        i = req
        while i < len(params):
            if params[i] is None:
                params[i] = defaults[i - req] if i - req < len(defaults) else None
            i += 1
        return params

    def get_int_params(self, text, index, num, req, *defaults):
        param_string = ''
        commas = 0
        while index < len(text) and text[index] in '$0123456789abcdefABCDEF,':
            char = text[index]
            if char == ',':
                commas += 1
                if commas == num:
                    break
            param_string += char
            index += 1
        params = self._get_params(param_string, num, req, defaults, True)
        return [index] + params

    def get_str_params(self, text, index, link_text='', chars=''):
        start = index
        valid_chars = '#' + chars
        while index < len(text) and (text[index] in valid_chars or text[index].isalnum()):
            index += 1
        param = text[start:index]
        end = index
        if index < len(text) and text[index] == '(':
            end = text.index(')', index) + 1
            link_text = text[index + 1:end - 1]
        return end, param, link_text

    def get_img_fname(self, fname, path):
        if fname:
            suffix = '.png'
            if fname[-len(suffix):].lower() == suffix:
                suffix = ''
            if fname[0] == '/':
                return '%s%s' % (fname[1:], suffix)
            return join(path, '%s%s' % (fname, suffix))

    def get_img_params(self, text, index, num, req, defaults, path, def_fname='', chars=''):
        valid_chars = '$0123456789abcdefABCDEF,' + chars
        param_string = ''
        commas = 0
        while index < len(text) and text[index] in valid_chars:
            char = text[index]
            if char == ',':
                commas += 1
                if commas == num:
                    break
            param_string += char
            index += 1
        params = self._get_params(param_string, num, req, defaults)
        if index < len(text) and text[index] == '(':
            end = text.index(')', index) + 1
            fname = text[index + 1:end - 1]
        else:
            end = index
            fname = def_fname
        img_fname = self.get_img_fname(fname, path)
        return [end, fname, img_fname] + params

    def get_udg_params(self, text, index, num, req, defaults, path, mask_step_index=3, chars=''):
        udg_params = self.get_img_params(text, index, num, req, defaults, path, chars=chars)
        end = udg_params[0]
        if end < len(text) and text[end] == ':':
            mask_defaults = (udg_params[mask_step_index + 3],)
            end, fname, img_fname, mask_addr, mask_step = self.get_img_params(text, end + 1, 2, 1, mask_defaults, path)
            udg_params[0:3] = [end, fname, img_fname]
            udg_params += [mask_addr, mask_step]
        else:
            udg_params += [None, None]
        return udg_params

    def get_numeric_param(self, text, index):
        num_start = index
        end = index
        digits = '0123456789'
        if text[index] == '$':
            num_start += 1
            end += 1
            digits += 'abcdefABCDEF'
        while end < len(text) and text[end] in digits:
            end += 1
        num_str = text[num_start:end]
        num = parse_int(text[index:end])
        return end, num_str, num

    def expand_bug(self, text, index, cwd):
        # #BUG[#name][(link text)]
        end, anchor, link_text = self.get_str_params(text, index, 'bug')
        bugs_file = FileInfo.relpath(cwd, self.paths[P_BUGS])
        link = '<a class="link" href="%s%s">%s</a>' % (bugs_file, anchor, link_text)
        return end, link

    def expand_c(self, text, index, cwd):
        # #C
        return index, '&#169;'

    def expand_call(self, text, index, cwd):
        # #CALL:methodName(args)
        macro = '#CALL'
        if text[index] != ':':
            self.error("Cannot parse malformed %s macro: %s%s..." % (macro, macro, text[index]))
        end, method_name, arg_string = self.get_str_params(text, index + 1, None, chars='_')
        if not hasattr(self, method_name) or not callable(getattr(self, method_name)):
            self.error("Unknown method name in %s macro: %s" % (macro, method_name))
        method = getattr(self, method_name)
        if arg_string is None:
            self.error("No argument list specified in %s macro: %s%s" % (macro, macro, text[index:end]))
        args = [cwd] + self._get_params(arg_string)
        return end, method(*args)

    def expand_desc(self, text, index, cwd):
        # #Daddress
        end, address = self.get_int_params(text, index, 1, 1)
        entry = self.skool.get_entry(address)
        desc = address
        if entry:
            if entry.description:
                desc = entry.description
            else:
                self.error('Entry at %s has no description' % address)
        else:
            self.error('Cannot determine description for nonexistent entry at %s' % address)
        return end, str(desc)

    def expand_erefs(self, text, index, cwd):
        # #EREFSaddress
        end, address = self.get_int_params(text, index, 1, 1)
        ereferrers = self.skool.get_entry_point_refs(address)
        addr_fmt = '%04X' if text[index] == '$' else '%i'
        if not ereferrers:
            self.error('Entry point at %s has no referrers' % (addr_fmt % address))
        ereferrers.sort()
        eref_files = [(FileInfo.asm_relpath(cwd, addr, self.skool.code_path), addr) for addr in ereferrers]
        html = 'routine at '
        if len(ereferrers) > 1:
            html = 'routines at '
            html += ', '.join('<a class="link" href="%s">%s</a>' % (eref_file, addr_fmt % addr) for eref_file, addr in eref_files[:-1])
            html += ' and '
        eref_file, addr = eref_files[-1]
        html += '<a class="link" href="%s">%s</a>' % (eref_file, addr_fmt % addr)
        return end, html

    def expand_fact(self, text, index, cwd):
        # #FACT[#name][(link text)]
        end, anchor, link_text = self.get_str_params(text, index, 'fact')
        facts_file = FileInfo.relpath(cwd, self.paths[P_FACTS])
        link = '<a class="link" href="%s%s">%s</a>' % (facts_file, anchor, link_text)
        return end, link

    def expand_font(self, text, index, cwd):
        # #FONTaddress,chars[,attr[,scale[,ftype]]][(fname)]
        end, fname, font_fname, address, chars, attr, scale, ftype = self.get_img_params(text, index, 5, 2, (56, 2, 0), self.paths['FontImagePath'], 'font')
        if self.need_image(font_fname):
            if ftype == 0:
                font_udg_array = [[Udg(attr, self.skool.snapshot[a:a + 8]) for a in range(address, address + 8 * chars, 8)]]
                width = 8 * chars
            else:
                char_widths = self.skool.snapshot[address:address + chars]
                width = 1 + sum(char_widths) + chars
                first_char = address % 256
                udgs = self.skool.get_text(''.join([chr(c) for c in range(first_char, first_char + chars)]))
                font_udg_array = [[Udg(attr, udg) for udg in udgs]]
            self.write_image(font_udg_array, font_fname, scale=scale, width=width)
        img_path = FileInfo.relpath(cwd, font_fname)
        return end, '<img alt="%s" src="%s" />' % (fname, img_path)

    def expand_link(self, text, index, cwd):
        # #LINK:PageId[#name](link text)
        macro = '#LINK'
        if text[index] != ':':
            self.error("Cannot parse malformed %s macro: %s%s..." % (macro, macro, text[index]))
        end, page_id, link_text = self.get_str_params(text, index + 1)
        anchor = ''
        if '#' in page_id:
            page_id, anchor = page_id.split('#')
            anchor = '#' + anchor
        if page_id not in self.paths:
            self.error("Unknown page ID in %s macro: %s" % (macro, page_id))
        href = FileInfo.relpath(cwd, self.paths[page_id])
        if not link_text:
            self.error("No link text specified in %s macro: %s%s" % (macro, macro, text[index:end]))
        link = '<a class="link" href="%s%s">%s</a>' % (href, anchor, link_text)
        return end, link

    def expand_name(self, text, index, cwd):
        # #NAMEname
        end, name, link_text = self.get_str_params(text, index)
        return end, '<a name="%s"></a>' % name

    def expand_p(self, text, index, cwd):
        # #P
        return index, '<br /><br />'

    def expand_poke(self, text, index, cwd):
        # #POKE[#name][(link text)]
        end, anchor, link_text = self.get_str_params(text, index, 'poke')
        pokes_file = FileInfo.relpath(cwd, self.paths[P_POKES])
        link = '<a class="link" href="%s%s">%s</a>' % (pokes_file, anchor, link_text)
        return end, link

    def expand_pokes(self, text, index, cwd):
        # #POKESaddress,byte[,length[,step]]
        end, address, byte, length, step = self.get_int_params(text, index, 4, 2, 1, 1)
        self.skool.snapshot[address:address + length * step:step] = [byte] * length
        return end, ''

    def expand_pops(self, text, index, cwd):
        # #POPS
        self.skool.pop_snapshot()
        return index, ''

    def expand_pushs(self, text, index, cwd):
        # #PUSHSname
        end, name, link_text = self.get_str_params(text, index)
        self.skool.push_snapshot(name)
        return end, ''

    def expand_refs(self, text, index, cwd):
        # #REFSaddress[(prefix)]
        end, addr_str, address = self.get_numeric_param(text, index)
        end, _, prefix = self.get_str_params(text, end, '')
        entry = self.skool.entries.get(address)
        if not entry:
            self.error('Cannot determine referrers for nonexistent entry at %s' % addr_str)
        addr_fmt = '%04X' if text[index] == '$' else '%i'
        referrers = [ref.address for ref in entry.referrers]
        if referrers:
            referrers.sort()
            ref_files = [(FileInfo.asm_relpath(cwd, addr, self.skool.code_path), addr) for addr in referrers]
            html = ('%s routine at ' % prefix).lstrip()
            if len(referrers) > 1:
                html = ('%s routines at ' % prefix).lstrip()
                html += ', '.join('<a class="link" href="%s">%s</a>' % (ref_file, addr_fmt % addr) for ref_file, addr in ref_files[:-1])
                html += ' and '
            ref_file, addr = ref_files[-1]
            html += '<a class="link" href="%s">%s</a>' % (ref_file, addr_fmt % addr)
        else:
            html = 'Not used directly by any other routines'
        return end, html

    def expand_reference(self, text, index, cwd):
        # #Raddress[@code][#anchor][(link text)]
        end, addr_str, address = self.get_numeric_param(text, index)
        container = self.skool.get_container(address)
        code_id = ''
        if text[end:].startswith('@'):
            end += 1
            while end < len(text) and text[end].isalnum():
                code_id += text[end]
                end += 1
        if not code_id and not container:
            self.error('Could not find routine file containing %s' % addr_str)
        if code_id:
            inst_addr_str = addr_str
        else:
            inst_addr_str = self.skool.get_instruction_addr_str(address)
        end, anchor, link_text = self.get_str_params(text, end, inst_addr_str)
        if code_id:
            container_address = address
            code_path = self.skool.get_code_path(code_id)
        else:
            container_address = container.address
            code_path = self.skool.code_path
            anchor = anchor if address == container_address else '#%s' % address
        ref_file = FileInfo.asm_relpath(cwd, container_address, code_path)
        link = '<a class="link" href="%s%s">%s</a>' % (ref_file, anchor, link_text)
        return end, link

    def expand_reg(self, text, index, cwd):
        # #REGx[y][']
        end = index
        while end < len(text) and text[end] in "abcdefhlirspxy'":
            end += 1
        reg = text[index:end]
        if not reg:
            self.error('#REG requires a register argument')
        elif len(reg) > 3:
            self.error('Bad register: "#REG%s"' % reg)
        reg_name = reg.lower() if self.case == CASE_LOWER else reg.upper()
        return end, '<span class="register">%s</span>' % reg_name

    def expand_scr(self, text, index, cwd):
        # #SCR[scale[,x[,y[,w[,h]]]]][(fname)]
        end, fname, scr_fname, scale, x, y, w, h = self.get_img_params(text, index, 5, 0, (1, 0, 0, 32, 24), self.paths['ScreenshotImagePath'], 'scr')
        if self.need_image(scr_fname):
            self.write_image(self.skool.get_screenshot(x, y, w, h), scr_fname, scale)
        img_path = FileInfo.relpath(cwd, scr_fname)
        return end, '<img alt="%s" src="%s" />' % (fname, img_path)

    def expand_space(self, text, index, cwd):
        # #SPACE[num]
        end, num_sp = self.get_int_params(text, index, 1, 0, 1)
        return end, '&#160;' * num_sp

    def expand_table(self, text, index, cwd):
        # #TABLE[(class[,col1class[,col2class...]])]<rows>TABLE#
        end, table = self.table_parser.parse_text(text, index)
        return end, self.build_table(table)

    def expand_udg(self, text, index, cwd):
        # #UDGaddr,attr[,scale,step,inc][:maskAddr[,maskStep]][(fname)]
        end, fname, udg_fname, addr, attr, scale, step, inc, mask_addr, mask_step = self.get_udg_params(text, index, 5, 2, (4, 1, 0), self.paths['UDGImagePath'])
        if not fname:
            fname = 'udg%s_%sx%s' % (addr, attr, scale)
            udg_fname = join(self.paths['UDGImagePath'], '%s.png' % fname)
        if self.need_image(udg_fname):
            udg_bytes = [self.skool.snapshot[addr + n * step] + inc for n in range(8)]
            mask_bytes = None
            mask = mask_addr is not None
            if mask:
                mask_bytes = self.skool.snapshot[mask_addr:mask_addr + 8 * mask_step:mask_step]
            self.write_image([[Udg(attr, udg_bytes, mask_bytes)]], udg_fname, scale, mask=mask)
        img_path = FileInfo.relpath(cwd, udg_fname)
        return end, '<img alt="%s" src="%s" />' % (fname, img_path)

    def expand_udgarray(self, text, index, cwd):
        # #UDGARRAYwidth[,attr[,scale[,step[,inc]]]];addr1[,attr1[,step1[,inc1]]];...(fname)
        udg_path = self.paths['UDGImagePath']
        end, fname, img_fname, width, def_attr, scale, def_step, def_inc = self.get_img_params(text, index, 5, 1, (56, 2, 1, 0), udg_path)
        udg_array = [[]]
        while end < len(text) and text[end] in ';':
            end, fname, img_fname, address, attr, step, inc = self.get_img_params(text, end + 1, 4, 1, (def_attr, def_step, def_inc), udg_path, chars='-x')
            num = 1
            if type(address) is int:
                addresses = [address]
            else:
                if 'x' in address:
                    address, num = address.split('x')
                    num = parse_int(num)
                if '-' in address:
                    elements = [parse_int(n) for n in address.split('-')]
                    if len(elements) < 3:
                        elements.append(1)
                    start_addr, end_addr, addr_step = elements
                    addresses = list(range(start_addr, end_addr + 1, addr_step))
                else:
                    addresses = [parse_int(address)]
            for a in addresses * num:
                udg_bytes = [self.skool.snapshot[a + n * step] + inc for n in range(8)]
                udg = Udg(attr, udg_bytes)
                if len(udg_array[-1]) == width:
                    udg_array.append([udg])
                else:
                    udg_array[-1].append(udg)
        if not fname:
            self.error('Filename missing in #UDGARRAY macro: #UDGARRAY%s' % text[index:end])
        if self.need_image(img_fname):
            self.write_image(udg_array, img_fname, scale)
        img_path = FileInfo.relpath(cwd, img_fname)
        return end, '<img alt="%s" src="%s" />' % (fname, img_path)

    def expand(self, text, cwd, **kwargs):
        if not text or text.find('#') < 0:
            return text

        while True:
            search = re.search('#[A-Z]+', text)
            if not search:
                break
            marker = search.group()
            if not marker in self.macros:
                self.error('Found unknown macro: %s' % marker)
            repf = self.macros[marker]
            if repf is None:
                self.error('Found unsupported macro: %s' % marker)
            start, index = search.span()
            end, rep = repf(text, index, cwd, **kwargs)
            text = text.replace(text[start:end], rep, 1)

        return text

class SkoolDazeWriter(SkoolWriter):
    def has_lessons(self):
        return True

    def keypress_table(self, cwd):
        keypress_routines = self.skool.keypress_routines
        snapshot = self.skool.snapshot

        table = Table('keytable', 'Code', 'Key', 'Address', 'Offset', 'Lookup', 'Keycode', 'Routine', 'Purpose')
        for c in range(32, 128):
            key = '&#169;' if c == 127 else cgi.escape(chr(c))
            address = c + 26624
            offset = snapshot[address]
            lookup = offset + 26624
            key_byte = snapshot[lookup]
            keycode = 'SPACE' if key_byte == 32 else chr(key_byte)
            routine = snapshot[lookup + 1] + 256 * snapshot[lookup + 2]
            purpose = keypress_routines[routine]
            routine_link = self.expand('#R%i' % routine, cwd)
            row = []
            row.append(Cell(c, "keyIndex"))
            row.append(Cell(key, "key"))
            row.append(Cell(address, "keyAddress"))
            row.append(Cell(offset, "data"))
            row.append(Cell(lookup, "keyAddress"))
            row.append(Cell(keycode, "centre"))
            row.append(Cell(routine_link, "keyAddress"))
            row.append(Cell(purpose, "purpose"))
            table.add_row(row)

        return table.html()

    def get_playarea_tile_text(self, tile):
        text = ['<i>ref_addr</i>=(x,y+152)=%i' % tile.ref_addr]
        text.append('<i>ref</i>=%i, <i>udg_page</i>=%i' % (tile.ref, tile.udg_page))
        text.append('<i>udg_addr</i>=(<i>ref</i>,<i>udg_page</i>)=%i' % (tile.ref + 256 * tile.udg_page))
        text.append('<i>attr_addr</i>=(x+128,y+152)=%i' % tile.attr_addr)
        text.append('<i>attr</i>=%i' % tile.attr)
        return text

    def _animatory_state_row(self, cwd, rownum):
        lines = []
        for state in (rownum, rownum + 128):
            lines.append('<td>')
            lines.append('<a name="%i" />%s' % (state, self.as_img(cwd, state, 2, 1, 120, 185)))
            lines.append('</td>')
        for state in (rownum, rownum + 128):
            lines.append('<td>')
            lines.append('<a name="%i" />%s' % (state, self.as_img(cwd, state, 2, 1, 120, 201)))
            lines.append('</td>')
        lines.append('<td>%i, %i: %s</td>' % (rownum, rownum + 128, self.skool.as_descs.get(rownum, '-')))
        return lines

    def _get_animatory_state_tiles_row(self, n):
        states = ((n, 185), (n + 128, 185), (n, 201), (n + 128, 201))
        states_desc = '%i, %i: %s' % (n, n + 128, self.skool.as_descs[n])
        return states, states_desc

class BackToSkoolWriter(SkoolWriter):
    def has_lessons(self):
        return True

    def write_mutable_tile_tables(self, cwd):
        lines = []

        doors = []
        doors.append((56637, True, 'Left study door'))
        doors.append((57088, True, 'Right study door'))
        doors.append((56576, True, 'Science Lab storeroom door'))
        doors.append((57149, False, "Boys' skool door"))
        doors.append((63744, False, 'Skool gate'))
        doors.append((56064, True, 'Drinks cabinet door'))

        windows = []
        windows.append((55040, True, 'Top floor window'))
        windows.append((55552, True, 'Middle floor window'))

        bike = [(57600, False, 'Tree without/with bike')]

        sections = (('Doors', doors), ('Windows (open)', windows), ('Tree (without bike)', bike))
        for title, section in sections:
            lines.append('<h3>%s</h3>\n<table class="mgrid">\n<tr>\n' % title)
            for a, initial_state, desc in section:
                addr = a + 256 if initial_state else a
                mutable_udgs = self.skool.get_mutable_udg_array(addr, -1)
                lines.append('<td>\n')
                lines += self.write_playarea_tile_table(cwd, mutable_udgs, 'm')
                lines.append('</td>\n')
            lines.append('</tr>\n</table>\n')

        cups = []
        cups.append((56694, "Leftmost cup on the shelf in the boys' skool (water/sherry)"))
        cups.append((56950, "Middle cup on the shelf in the boys' skool (water/sherry)"))
        cups.append((57206, "Rightmost cup on the shelf in the boys' skool (water/sherry)"))
        cups.append((57462, "Cup on the shelf in the girls' skool (water/sherry)"))
        lines.append('<h3>Cups</h3>\n<table>\n')
        for a, desc, in cups:
            lines.append('<tr>\n<td>\n<table class="mgrid">\n<tr>\n')
            for address in (a, a + 5):
                cup_udg_array = self.skool.get_mutable_udg_array(address, -1)
                lines.append('<td>\n')
                lines += self.write_playarea_tile_table(cwd, cup_udg_array, 'c%i_' % address)
                lines.append('</td>\n')
            lines.append('</tr>\n</table>\n</td><td>%s</td></tr>\n' % desc)
        lines.append('</table>\n')

        return lines

    def keypress_table(self, cwd):
        keypress_routines = self.skool.keypress_routines
        snapshot = self.skool.snapshot

        table = Table('keytable', 'Index', 'Key', 'Address', 'Offset', 'Lookup', 'Routine', 'Purpose')
        for index in range(80):
            key = '&#169;' if index == 79 else cgi.escape(chr(index + 48))
            address = index + 58624
            offset = snapshot[address]
            lookup = routine_link = purpose = ''
            if offset:
                lookup = offset + 58624
                routine = snapshot[lookup] + 256 * snapshot[lookup + 1]
                purpose = keypress_routines[routine]
                routine_link = self.expand('#R%i' % routine, cwd)
            row = []
            row.append(Cell(index, "keyIndex"))
            row.append(Cell(key, "key"))
            row.append(Cell(address, "keyAddress"))
            row.append(Cell(offset, "data"))
            row.append(Cell(lookup, "keyAddress"))
            row.append(Cell(routine_link, "keyAddress"))
            row.append(Cell(purpose, "purpose"))
            table.add_row(row)

        return table.html()

    def get_playarea_tile_text(self, tile):
        text = ['<i>Q_addr</i>=(x,181)=%i, <i>Q</i>=%i' % (tile.q_addr, self.skool.snapshot[tile.q_addr])]
        text.append('<i>ref_addr</i>=(<i>Q</i>,y+160)=%i' % tile.ref_addr)
        text.append('<i>ref</i>=%i, <i>udg_page</i>=%i' % (tile.ref, tile.udg_page))
        text.append('<i>udg_addr</i>=(<i>ref</i>,<i>udg_page</i>)=%i' % (tile.ref + 256 * tile.udg_page))
        text.append('<i>attr</i>=%i' % tile.attr)
        return text

    def mutable(self, cwd, address, padding=0):
        fname = 'mutable%i_%i' % (address, padding) if padding else 'mutable%i' % address
        mutable_fname = join('images', 'mutables', '%s.png' % fname)
        if self.need_image(mutable_fname):
            mutable_udgs = self.skool.get_mutable_udg_array(address, padding)
            self.write_image(mutable_udgs, mutable_fname, 2)
        img_path = FileInfo.relpath(cwd, mutable_fname)
        return '<img alt="%s" src="%s" />' % (fname, img_path)

class ContactSamCruiseWriter(SkoolWriter):
    def keypress_table(self, cwd):
        keypress_routines = self.skool.keypress_routines
        snapshot = self.skool.snapshot

        table = Table('keytable', 'Index', 'Key', 'Address', 'Offset', 'Lookup', 'Routine', 'Purpose')
        for index in range(80):
            key = '&#169;' if index == 79 else cgi.escape(chr(index + 48))
            address = index + 49968
            offset = snapshot[address]
            lookup = routine_link = purpose = ''
            if offset:
                lookup = offset + 60671
                routine = snapshot[lookup] + 256 * snapshot[lookup + 1]
                purpose = keypress_routines[routine]
                routine_link = self.expand('#R%i' % routine, cwd)
            row = []
            row.append(Cell(index, "keyIndex"))
            row.append(Cell(key, "key"))
            row.append(Cell(address, "keyAddress"))
            row.append(Cell(offset, "data"))
            row.append(Cell(lookup, "keyAddress"))
            row.append(Cell(routine_link, "keyAddress"))
            row.append(Cell(purpose, "purpose"))
            table.add_row(row)

        return table.html()

    def command_lists(self, cwd):
        lines = ['<table class="map">']
        lines.append('<tr><th>Address</th><th>Description</th></tr>')
        tap_descs = self.skool.tap_descs
        tap_addresses = tap_descs.keys()
        tap_addresses.sort()
        for tap_address in tap_addresses:
            lines.append(self.expand('<tr><td class="data">#R%i</td><td>%s</td></tr>' % (tap_address, tap_descs[tap_address]), cwd))
        lines.append('</table>')
        return '\n'.join(lines)

def join(*path_components):
    return '/'.join([c for c in path_components if c.replace('/', '')])

class FileInfo:
    def __init__(self, topdir, game_dir, replace_images):
        self.odir = join(topdir, game_dir)
        self.replace_images = replace_images

    def open_file(self, *names):
        path = self.odir
        for name in names:
            path = join(path, name)
        if not isdir(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))
        return open(path, 'wb')

    def need_image(self, fname):
        return self.replace_images or not self.file_exists(fname)

    def file_exists(self, fname):
        return isfile(join(self.odir, fname))

    @classmethod
    def relpath(cls, cwd, target):
        # Use os.path.relpath(target, cwd) in Python 2.6
        while cwd and cwd[-1] == '/':
            cwd = cwd[:-1]
        cwddirs = cwd.split('/')
        while cwddirs and not cwddirs[0]:
            cwddirs.pop(0)
        targetdirs = target.split('/')
        while cwddirs and cwddirs[0] == targetdirs[0]:
            cwddirs.pop(0)
            targetdirs.pop(0)
        return posixpath.normpath(('../' * len(cwddirs)) + '/'.join(targetdirs))

    @classmethod
    def asm_fname(cls, address, path=''):
        return posixpath.normpath(join(path, '%s.html' % address))

    @classmethod
    def lesson_fname(cls, lesson_num):
        return '%s.html' % lesson_num

    @classmethod
    def asm_relpath(cls, cwd, address, path):
        return cls.relpath(cwd, join(path, cls.asm_fname(address)))

class Table:
    def __init__(self, cls='', *headers):
        self.cls = cls
        self.width = None
        self.rows = []
        if headers:
            self.rows.append([Header(h) for h in headers])
        self.cellpadding = None

    def add_row(self, row):
        self.rows.append(row)

    def html(self):
        lines = []
        cls_attr = ' class="%s"' % self.cls if self.cls else ''
        width_attr = ' width="%i"' % self.width if self.width else ''
        cp_attr = ' cellpadding="%i"' % self.cellpadding if self.cellpadding is not None else ''
        lines.append('<table%s%s%s>' % (width_attr, cls_attr, cp_attr))
        for row in self.rows:
            lines.append('<tr>')
            for cell in row:
                lines.append(cell.html())
            lines.append('</tr>')
        lines.append('</table>\n')
        return '\n'.join(lines)

class Cell:
    def __init__(self, contents, cls='', colspan=1, rowspan=1, tag='td'):
        self.contents = str(contents)
        self.cls = cls
        self.colspan = colspan
        self.rowspan = rowspan
        self.tag = tag

    def html(self):
        cls_attr = ' class="%s"' % self.cls if self.cls else ''
        cs_attr = ' colspan="%s"' % self.colspan if self.colspan > 1 else ''
        rs_attr = ' rowspan="%s"' % self.rowspan if self.rowspan > 1 else ''
        if self.contents:
            return '<%s%s%s%s>%s</%s>' % (self.tag, cls_attr, cs_attr, rs_attr, self.contents, self.tag)
        return '<%s%s%s%s/>' % (self.tag, cls_attr, cs_attr, rs_attr)

class Header(Cell):
    def __init__(self, contents, cls='', colspan=1, rowspan=1):
        Cell.__init__(self, contents, cls, colspan, rowspan, 'th')

def get_skool_writer_class(name):
    if name == 'SkoolDazeWriter':
        return SkoolDazeWriter
    if name == 'BackToSkoolWriter':
        return BackToSkoolWriter
    if name == 'ContactSamCruiseWriter':
        return ContactSamCruiseWriter
    return SkoolWriter
