# -*- coding: utf-8 -*-
import unittest
from skoolkittest import SkoolKitTestCase

TEST_SKOOL = 'test-asm.skool'

USAGE = """Usage: skool2asm.py [options] <skoolfile>

Options:
  -q   Be quiet (don't show timings)
  -v   Be verbose (show instruction substitutions)
  -w   Suppress warnings
  -e   Write processed lines to stderr (implies -w)
  -d   Use CR+LF to end lines
  -t   Use tab to indent instructions (default indentation is 2 spaces)
  -l   Write disassembly in lower case
  -u   Write disassembly in upper case
  -D   Write disassembly in decimal
  -H   Write disassembly in hexadecimal
  -iN  Set instruction field width to N (default=23)
  -fN  Apply fixes:
         N=0: None (default)
         N=1: @ofix only
         N=2: @ofix and @bfix
         N=3: @ofix, @bfix and @rfix (implies -r)
  -c   Create default labels for unlabelled instructions
  -s   Use safe substitutions (@ssub)
  -r   Use relocatability substitutions too (@rsub) (implies -f1)"""

class Skool2AsmTest(SkoolKitTestCase):
    def test_no_arguments(self):
        cmdline, output, error = self.run_skool2asm(err_lines=True)
        self.assert_output_equal(error, USAGE.split('\n'), False)

    def test_no_options(self):
        asm = self.get_asm(TEST_SKOOL)
        self.assertEqual(asm[0], '  ORG 24576') # no crlf, tabs or lower case
        self.assertEqual(asm[33], '  LD A,B')   # @ofix
        self.assertEqual(asm[34], '  LD B,A')   # @ofix block
        self.assertEqual(asm[38], '  LD C,A')   # @bfix
        self.assertEqual(asm[39], '  LD D,A')   # @bfix block
        self.assertEqual(asm[43], '  LD E,A')   # @rfix block
        self.assertEqual(asm[46], '  RET')      # @ssub
        self.assertEqual(asm[50], '  LD A,0')   # @rsub

    def test_d(self):
        asm = self.get_asm('-d %s' % TEST_SKOOL, strip_cr=False)
        self.assertEqual(asm[0][-1], '\r')

    def test_t(self):
        asm = self.get_asm('-t %s' % TEST_SKOOL)
        self.assertEqual(asm[0], '\tORG 24576')
        self.assertEqual(asm[9][:12], '\tLD HL,$6003')

    def test_l(self):
        asm = self.get_asm('-l %s' % TEST_SKOOL)
        self.assertEqual(asm[8], 'DOSTUFF:') # Labels unaffected
        self.assertEqual(asm[19][:5], '  nop')

    def test_i(self):
        width = 30
        asm = self.get_asm('-i%i %s' % (width, TEST_SKOOL))
        self.assertEqual(asm[9].find(';'), width + 3)

    def test_f0(self):
        asm = self.get_asm('-f0 %s' % TEST_SKOOL)
        self.assertEqual(asm[33], '  LD A,B')
        self.assertEqual(asm[34], '  LD B,A')
        self.assertEqual(asm[38], '  LD C,A')
        self.assertEqual(asm[39], '  LD D,A')
        self.assertEqual(asm[43], '  LD E,A')

    def test_f1(self):
        asm = self.get_asm('-f1 %s' % TEST_SKOOL)
        self.assertEqual(asm[33], '  LD A,C')
        self.assertEqual(asm[34], '  LD B,C')
        self.assertEqual(asm[38], '  LD C,A')
        self.assertEqual(asm[39], '  LD D,A')
        self.assertEqual(asm[43], '  LD E,A')

    def test_f2(self):
        asm = self.get_asm('-f2 %s' % TEST_SKOOL)
        self.assertEqual(asm[33], '  LD A,C')
        self.assertEqual(asm[34], '  LD B,C')
        self.assertEqual(asm[38], '  LD C,B')
        self.assertEqual(asm[39], '  LD D,B')
        self.assertEqual(asm[43], '  LD E,A')

    def test_f3(self):
        asm = self.get_asm('-f3 %s' % TEST_SKOOL)
        self.assertEqual(asm[33], '  LD A,C')
        self.assertEqual(asm[34], '  LD B,C')
        self.assertEqual(asm[38], '  LD C,B')
        self.assertEqual(asm[39], '  LD D,B')
        self.assertEqual(asm[43], '  LD E,B')

    def test_s(self):
        asm = self.get_asm('-s %s' % TEST_SKOOL)
        self.assertEqual(asm[46], '  JP (HL)')

    def test_r(self):
        asm = self.get_asm('-r %s' % TEST_SKOOL)
        self.assertEqual(asm[33], '  LD A,C')
        self.assertEqual(asm[34], '  LD B,C')
        self.assertEqual(asm[50], '  XOR A')

    def test_no_html_escape(self):
        asm = self.get_asm(TEST_SKOOL)
        self.assertEquals(asm[9], '  LD HL,$6003             ; a <= b and b >= c')

    def test_macro_reduction(self):
        asm = self.get_asm(TEST_SKOOL, out_lines=False)
        self.assertEqual(asm.count('#'), 1) # Only one '#' (at 24582)
        asm = [line.rstrip() for line in asm.split('\n', 16)]
        self.assertEqual(asm[14], '  DEFB 123                ; A=0')
        self.assertEqual(asm[15], '  DEFB 246                ; DOSTUFF')

    def get_asm(self, args, out_lines=True, err_lines=True, strip_cr=True):
        cmdline, output, error = self.run_skool2asm(args, out_lines, err_lines, strip_cr)
        self.assertEqual('Wrote ASM to stdout', error[-1][:19], 'Error(s) while running "%s"' % cmdline)
        return output

if __name__ == '__main__':
    unittest.main()
