# -*- coding: utf-8 -*-
from os import remove
import unittest
from skoolkittest import SkoolKitTestCase

TEST_SNA = 'test-sna.sna'
TEST_CTL = 'test-sna-ctl.ctl'

USAGE = """Usage: sna2skool.py [options] file.[sna|z80|bin]

Options:
  -c FILE  Use FILE as the ctl file (default is file.ctl)
  -g FILE  Generate a control file in FILE
  -h       Write hexadecimal addresses in the generated control file
  -H       Write hexadecimal addresses and operands in the disassembly
  -L       Write the disassembly in lower case
  -s ADDR  Specify the address at which to start disassembling
  -o ADDR  Specify the origin address of file.bin (default: 65536 - length)
  -t       Show ASCII text in the comment fields
  -r       Suppress addition of comments listing entry point referrers
  -n N     Set the max number of bytes per DEFB statement to N (default=8)
  -m M     Group DEFB blocks by addresses that divide M
  -z       Write bytes with leading zeroes in DEFB statements
  -l L     Set the max number of characters per DEFM statement to L (default=66)"""

class OptionsTest(SkoolKitTestCase):
    def test_no_arguments(self):
        cmdline, output, error = self.run_sna2skool(err_lines=True)
        self.assert_output_equal(error, USAGE.split('\n'), False)

    def test_no_options(self):
        lines = self.write_skool(TEST_SNA, 10)
        self.assertEqual(lines[0], '; @start')
        self.assertEqual(lines[1], '; @org=16384')
        self.assertEqual(lines[3], '; Routine at 16384')
        self.assertEqual(lines[4][:16], 'c16384 LD A,(DE)')

    def test_t(self):
        skool = self.write_skool('-t %s' % TEST_SNA)
        self.assertTrue('*23460 LD SP,12854   ; [162]' in skool)

    def test_s(self):
        start = '24832'
        lines = self.write_skool('-s %s %s' % (start, TEST_SNA), 10)
        self.assertTrue('; @org=24832' in lines)
        self.assertTrue(self.get_first_address(lines) == start)

    def test_sg(self):
        ctlfile = 'test_sg.ctl'
        self.write_skool('-s 23296 -g %s %s' % (ctlfile, TEST_SNA))
        self.assertTrue(self.files_match(TEST_CTL, ctlfile))
        remove(ctlfile)

    def test_c(self):
        lines = self.write_skool('-c %s %s' % (TEST_CTL, TEST_SNA), 10)
        self.assertTrue(self.get_first_address(lines) == '23296')

    def write_skool(self, args, split=0):
        cmdline, output, error = self.run_sna2skool(args, out_lines=False)
        self.assert_no_errors(error, cmdline)
        if split:
            return [line.rstrip() for line in output.split('\n', split)]
        return output

    def files_match(self, fname1, fname2):
        lines1 = self.get_lines(fname1)
        lines2 = self.get_lines(fname2)
        if len(lines2) == len(lines1):
            for i in range(len(lines1)):
                if lines2[i] != lines1[i]:
                    break
            else:
                return True

    def get_lines(self, fname):
        f = open(fname, 'r')
        lines = []
        for line in f:
            lines.append(line.rstrip())
        f.close()
        return lines

    def get_first_address(self, lines):
        for line in lines:
            if line.startswith('c'):
                return line[1:6]

if __name__ == '__main__':
    unittest.main()
